# -*- coding: utf-8 -*-

class ExceptionReactionSystem():
    class InvalidSyntax(Exception): pass
    class ReactantSetCannotBeEmpty(Exception): pass
    class ProductSetCannotBeEmpty(Exception): pass
    class InvalidReaction(Exception): pass
    class InvalidReactionSet(Exception): pass
    class InvalidNumber(Exception): pass
    class InvalidFormula(Exception): pass
    class InvalidContextSet(Exception): pass


import sys
import multiprocessing
from PyQt5 import QtWidgets
from fbp_calculator.mainwindowfbp import MainWindowFBP

def main():
    
    # https://github.com/pyinstaller/pyinstaller/wiki/Recipe-Multiprocessing
    multiprocessing.freeze_support()

    app = QtWidgets.QApplication(sys.argv)
    mainWindowFBP = MainWindowFBP(app)
    mainWindowFBP.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'formula_ui.ui'
#
# Created by: PyQt5 UI code generator 5.10.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_DialogFBP(object):
    def setupUi(self, DialogFBP):
        DialogFBP.setObjectName("DialogFBP")
        DialogFBP.resize(380, 400)
        self.gridLayout = QtWidgets.QGridLayout(DialogFBP)
        self.gridLayout.setObjectName("gridLayout")
        self.lineEditSymbols = QtWidgets.QLineEdit(DialogFBP)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lineEditSymbols.sizePolicy().hasHeightForWidth())
        self.lineEditSymbols.setSizePolicy(sizePolicy)
        self.lineEditSymbols.setFocusPolicy(QtCore.Qt.NoFocus)
        self.lineEditSymbols.setReadOnly(True)
        self.lineEditSymbols.setObjectName("lineEditSymbols")
        self.gridLayout.addWidget(self.lineEditSymbols, 2, 0, 1, 1)
        self.textBrowserFormula = QtWidgets.QTextBrowser(DialogFBP)
        self.textBrowserFormula.setEnabled(False)
        self.textBrowserFormula.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.textBrowserFormula.setObjectName("textBrowserFormula")
        self.gridLayout.addWidget(self.textBrowserFormula, 7, 0, 1, 2)
        self.labelSteps = QtWidgets.QLabel(DialogFBP)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.labelSteps.sizePolicy().hasHeightForWidth())
        self.labelSteps.setSizePolicy(sizePolicy)
        self.labelSteps.setMinimumSize(QtCore.QSize(50, 0))
        self.labelSteps.setObjectName("labelSteps")
        self.gridLayout.addWidget(self.labelSteps, 1, 1, 1, 1)
        self.gridLayout_2 = QtWidgets.QGridLayout()
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.labelLoadingImage = QtWidgets.QLabel(DialogFBP)
        self.labelLoadingImage.setText("")
        self.labelLoadingImage.setObjectName("labelLoadingImage")
        self.gridLayout_2.addWidget(self.labelLoadingImage, 0, 4, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout_2.addItem(spacerItem, 0, 2, 1, 1)
        self.comboBoxFormulaType = QtWidgets.QComboBox(DialogFBP)
        self.comboBoxFormulaType.setEnabled(False)
        self.comboBoxFormulaType.setObjectName("comboBoxFormulaType")
        self.comboBoxFormulaType.addItem("")
        self.comboBoxFormulaType.addItem("")
        self.comboBoxFormulaType.addItem("")
        self.gridLayout_2.addWidget(self.comboBoxFormulaType, 0, 1, 1, 1)
        self.labelComputing = QtWidgets.QLabel(DialogFBP)
        self.labelComputing.setObjectName("labelComputing")
        self.gridLayout_2.addWidget(self.labelComputing, 0, 3, 1, 1)
        self.toolButtonSave = QtWidgets.QToolButton(DialogFBP)
        icon = QtGui.QIcon.fromTheme("document-save-as")
        self.toolButtonSave.setIcon(icon)
        self.toolButtonSave.setIconSize(QtCore.QSize(18, 18))
        self.toolButtonSave.setAutoRaise(True)
        self.toolButtonSave.setObjectName("toolButtonSave")
        self.gridLayout_2.addWidget(self.toolButtonSave, 0, 5, 1, 1)
        self.gridLayout.addLayout(self.gridLayout_2, 5, 0, 1, 2)
        self.labelSymbols = QtWidgets.QLabel(DialogFBP)
        self.labelSymbols.setObjectName("labelSymbols")
        self.gridLayout.addWidget(self.labelSymbols, 1, 0, 1, 1)
        self.lineEditSteps = QtWidgets.QLineEdit(DialogFBP)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lineEditSteps.sizePolicy().hasHeightForWidth())
        self.lineEditSteps.setSizePolicy(sizePolicy)
        self.lineEditSteps.setMinimumSize(QtCore.QSize(50, 0))
        self.lineEditSteps.setFocusPolicy(QtCore.Qt.NoFocus)
        self.lineEditSteps.setReadOnly(True)
        self.lineEditSteps.setObjectName("lineEditSteps")
        self.gridLayout.addWidget(self.lineEditSteps, 2, 1, 1, 1)
        self.tableWidgetFormula = QtWidgets.QTableWidget(DialogFBP)
        self.tableWidgetFormula.setEnabled(False)
        self.tableWidgetFormula.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.tableWidgetFormula.setAlternatingRowColors(True)
        self.tableWidgetFormula.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)
        self.tableWidgetFormula.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.tableWidgetFormula.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.tableWidgetFormula.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.tableWidgetFormula.setGridStyle(QtCore.Qt.SolidLine)
        self.tableWidgetFormula.setWordWrap(False)
        self.tableWidgetFormula.setObjectName("tableWidgetFormula")
        self.tableWidgetFormula.setColumnCount(0)
        self.tableWidgetFormula.setRowCount(0)
        self.tableWidgetFormula.horizontalHeader().setCascadingSectionResizes(True)
        self.tableWidgetFormula.horizontalHeader().setDefaultSectionSize(0)
        self.tableWidgetFormula.horizontalHeader().setMinimumSectionSize(30)
        self.tableWidgetFormula.verticalHeader().setVisible(False)
        self.tableWidgetFormula.verticalHeader().setDefaultSectionSize(25)
        self.tableWidgetFormula.verticalHeader().setMinimumSectionSize(0)
        self.gridLayout.addWidget(self.tableWidgetFormula, 9, 0, 1, 2)
        self.listFormula = QtWidgets.QListWidget(DialogFBP)
        self.listFormula.setEnabled(False)
        self.listFormula.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.listFormula.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.listFormula.setAlternatingRowColors(True)
        self.listFormula.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)
        self.listFormula.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.listFormula.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.listFormula.setUniformItemSizes(False)
        self.listFormula.setObjectName("listFormula")
        self.gridLayout.addWidget(self.listFormula, 8, 0, 1, 2)
        self.line = QtWidgets.QFrame(DialogFBP)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.gridLayout.addWidget(self.line, 3, 0, 1, 2)

        self.retranslateUi(DialogFBP)
        QtCore.QMetaObject.connectSlotsByName(DialogFBP)
        DialogFBP.setTabOrder(self.lineEditSymbols, self.lineEditSteps)

    def retranslateUi(self, DialogFBP):
        _translate = QtCore.QCoreApplication.translate
        DialogFBP.setWindowTitle(_translate("DialogFBP", " "))
        self.labelSteps.setText(_translate("DialogFBP", "Steps:"))
        self.comboBoxFormulaType.setItemText(0, _translate("DialogFBP", "Formula"))
        self.comboBoxFormulaType.setItemText(1, _translate("DialogFBP", "List"))
        self.comboBoxFormulaType.setItemText(2, _translate("DialogFBP", "Table"))
        self.labelComputing.setText(_translate("DialogFBP", "Computing"))
        self.toolButtonSave.setText(_translate("DialogFBP", "Save As"))
        self.labelSymbols.setText(_translate("DialogFBP", "Symbols:"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    DialogFBP = QtWidgets.QDialog()
    ui = Ui_DialogFBP()
    ui.setupUi(DialogFBP)
    DialogFBP.show()
    sys.exit(app.exec_())




# -*- coding: utf-8 -*-

from fbp_calculator.reactionsystem.boolean_wrap import (Or, ZERO)
    
from fbp_calculator.reactionsystem.reaction import Reaction
from fbp_calculator.reactionsystem.exceptions import ExceptionReactionSystem


class ReactionSet(set):
    def __init__(self, reactions=None):
        if reactions is None:
            super(ReactionSet, self).__init__()
        else:
            for reaction in reactions:
                self.add(reaction)

    def add(self, reaction):
        if not isinstance(reaction, Reaction): raise ExceptionReactionSystem.InvalidReaction()
        super(ReactionSet, self).add(reaction)


    def cause(self, symbol):
        Reaction._check_symbol(symbol)

        cause = ZERO
        for reaction in self:
            if symbol in reaction.P:
                cause = Or(cause, reaction.ap())
        return cause




# -*- coding: utf-8 -*-

import re
from copy import deepcopy
from PyQt5 import QtCore, QtGui, QtWidgets
from fbp_calculator.reactionsystem import (
    Reaction,
    ReactionSet,
    ExceptionReactionSystem)
from fbp_calculator.ui_mainwindowfbp import Ui_MainWindowFBP
from fbp_calculator.dialogfbp import DialogFBP
from fbp_calculator.reaction_adapter import reaction_adapter, reaction_invadapter
from fbp_calculator import __version__


class MainWindowFBP(QtWidgets.QMainWindow, Ui_MainWindowFBP):
    def __init__(self, app, parent=None):
        super(MainWindowFBP, self).__init__(parent)

        self.setupUi(self)

        self._translate = QtCore.QCoreApplication.translate
 
        self.statusbarStyle = self.statusbar.styleSheet()

        self.reaction_list = []
        
        self.temp_file_name = ''
        self.current_file_name = ''

        self.validatorLineEditSymbols_regex = '^([a-zA-Z]|\d|\s|,)+$' # pylint: disable=W1401
        self.validatorLineEditSymbols = QtGui.QRegExpValidator(QtCore.QRegExp(self.validatorLineEditSymbols_regex))
        self.validatorLineEditSymbols_empty_regex = '^([a-zA-Z]|\d|\s|,)*$' # pylint: disable=W1401
        self.validatorLineEditSymbols_empty = QtGui.QRegExpValidator(QtCore.QRegExp(self.validatorLineEditSymbols_empty_regex))
        self.validatorContextProperties_regex = '^([a-zA-Z]|\d|\s|,|%)+$' # pylint: disable=W1401
        self.validatorContextProperties = QtGui.QRegExpValidator(QtCore.QRegExp(self.validatorContextProperties_regex))


        self.lineEditReactants.setValidator(self.validatorLineEditSymbols)
        self.lineEditProducts.setValidator(self.validatorLineEditSymbols)
        self.lineEditInhibitors.setValidator(self.validatorLineEditSymbols_empty)
        self.lineEditCalculatorSymbols.setValidator(self.validatorLineEditSymbols)

        self.statusbar.messageChanged.connect(self.statusbarChanged)

        self.pushButtonAdd.clicked.connect(self.pushButtonAdd_clicked)
        self.pushButtonDelete.clicked.connect(self.pushButtonDelete_clicked)
        self.pushButtonCalculate.clicked.connect(self.pushButtonCalculate_clicked)
        
        self.lineEditReactants.textChanged.connect(self.pushButtonAdd_enable)
        self.lineEditProducts.textChanged.connect(self.pushButtonAdd_enable)
        self.listWidgetReactions._checked_item_number = 0
        self.listWidgetReactions.itemChanged.connect(self.listWidgetReactions_itemChanged)
        self.lineEditCalculatorSymbols.textChanged.connect(self.pushButtonCalculate_enable)
        self.spinBoxCalculatorSteps.valueChanged.connect(self.pushButtonCalculate_enable)

        self.actionNew.triggered.connect(self.actionNew_triggered)
        self.actionOpen.triggered.connect(self.actionOpen_triggered)
        self.actionSave_as.triggered.connect(self.actionSave_as_triggered)
        self.actionSave.triggered.connect(self.actionSave_triggered)
        self.actionQuit.triggered.connect(self.actionQuit_triggered)

        self.actionAbout.triggered.connect(self.actionAbout_triggered)

        self.tableWidgetProperties.horizontalScrollBar().valueChanged.connect(self.tableWidgetProperties_scrollBar_valueChanged)
        self.tableWidgetProperties_addColumn()

        height = self.tableWidgetProperties.horizontalHeader().size().height()
        height += self.tableWidgetProperties.rowHeight(0)
        height += self.tableWidgetProperties.rowHeight(1)
        height += self.tableWidgetProperties.horizontalScrollBar().size().height()
        self.tableWidgetProperties.setFixedHeight(height+4)

        self.setGeometry(
            QtWidgets.QStyle.alignedRect(
                QtCore.Qt.LeftToRight,
                QtCore.Qt.AlignCenter,
                self.size(),
                app.desktop().availableGeometry()))


    def resizeEvent(self, event):
        self.tableWidgetProperties_fillSpace()

    def closeEvent(self, event):
        event.ignore()
        self.actionQuit_triggered()


    def actionNew_triggered(self, value=None):
        if not self.check_save():
            return
            
        self.spinBoxCalculatorSteps.setValue(1)
        self.tableWidgetProperties.cellWidget(0,0).setText('')
        self.tableWidgetProperties.cellWidget(1,0).setText('')
        self.lineEditCalculatorSymbols.setText('')
        self.lineEditReactants.setText('')
        self.lineEditProducts.setText('')
        self.lineEditInhibitors.setText('')
        self.lineEditReactants.setFocus(True)

        self.current_file_name = ''
        self.reaction_list.clear()
        self.listWidgetReactions_clear()

    def actionOpen_triggered(self, value=None):
        if not self.check_save():
            return
        self.temp_file_name, _ = \
            QtWidgets.QFileDialog.getOpenFileName(self, 
                'Open a Reaction System file', 
                '',
                'TXT files (*.txt);;All files (*)')
        if not self.temp_file_name:
            return

        try:
            with open(self.temp_file_name, 'r') as file:
                file_content = file.read()
        except Exception as e:
            error_message = str(e)
            if 'Errno' in error_message:
                error_message = error_message.split('] ')[1]
            QtWidgets.QMessageBox.critical(self,
                'Error when opening the file',
                '{}'.format(error_message),
                QtWidgets.QMessageBox.Close,
                QtWidgets.QMessageBox.Close)
            return

        try:
            self.reaction_list = self.text2reaction_list(file_content)
        except Exception as e:
            QtWidgets.QMessageBox.critical(self,
                'Error when opening the file',
                'Invalid syntax: \'{}\''.format(self.temp_file_name),
                QtWidgets.QMessageBox.Close,
                QtWidgets.QMessageBox.Close)
            return

        self.listWidgetReactions_clear()

        for reaction in self.reaction_list:
            self.listWidgetReactions_addReaction(reaction)

        self.current_file_name = self.temp_file_name

        self.notify('File ' + self.current_file_name + ' opened')

    def actionSave_triggered(self, value=None):
        if not self.current_file_name:
            self.actionSave_as_triggered()
            return
        self.save_file(self.current_file_name)
    
    def actionSave_as_triggered(self, value=None):
        file_name, _ = \
            QtWidgets.QFileDialog.getSaveFileName(self, 
                'Save Reaction System As', 
                'untitled.txt',
                'TXT files (*.txt)')
        self.save_file(file_name)

    def actionQuit_triggered(self, value=None):
        if not self.check_save():
            return
        QtCore.QCoreApplication.quit()

    def actionAbout_triggered(self):
        QtWidgets.QMessageBox.about(self,
            self._translate('MainWindowFBP', 'FBP Calculator'),
            'version {}\n'.format(__version__) +
            'Written by William Guglielmo.\n' + \
            'Based on the work of:\n' + \
            'Roberto Barbuti\n' + \
            'Roberta Gori\n' + \
            'Paolo Milazzo\n' + \
            'Francesca Levi\n')

    @staticmethod 
    def text2reaction_list(text):
        reaction_list = []
        text.replace('\r', ' ')
        reactions = text.split('\n')
        for reaction in reactions:
            if (re.match('^\s*#', reaction) or re.match('^\s*$', reaction)): # pylint: disable=W1401
                continue
            reaction_list.append(Reaction(reaction_adapter(reaction)))
        return reaction_list

    @staticmethod
    def reaction_list2text(reaction_list):
        text = ''
        for reaction in reaction_list:
            text += reaction_invadapter(str(reaction)) + '\r\n'
        return text


    def save_file(self, file_name):
        if not file_name:
            return
        try:
            with open(file_name, 'w') as file:
                file.write(self.reaction_list2text(self.reaction_list))
        except Exception as e:
            error_message = str(e)
            if 'Errno' in error_message:
                error_message = error_message.split('] ')[1]
            QtWidgets.QMessageBox.critical(self,
                'Error when saving the file',
                '{}'.format(error_message),
                QtWidgets.QMessageBox.Close,
                QtWidgets.QMessageBox.Close)
            return
        self.actionSave.setEnabled(False)
        self.current_file_name = file_name
        self.notify('File ' + self.current_file_name + ' saved')

    def check_save(self):
        if self.actionSave.isEnabled() and len(self.reaction_list):
            buttonReply = QtWidgets.QMessageBox.warning(self,
                'Save changes before closing?',
                'Your changes will be lost if you don’t save them.',
                QtWidgets.QMessageBox.Cancel | QtWidgets.QMessageBox.Discard | QtWidgets.QMessageBox.Save,
                QtWidgets.QMessageBox.Cancel)

            if buttonReply == QtWidgets.QMessageBox.Save:
                self.actionSave_triggered()
            elif buttonReply == QtWidgets.QMessageBox.Discard:
                pass
            else:
                return False
        return True

    
    def pushButtonAdd_clicked(self):
        reactants = self.lineEditReactants.text()
        products = self.lineEditProducts.text()
        inhibitors = self.lineEditInhibitors.text()

        try:
            reaction = Reaction(
                R=reaction_adapter(reactants),
                P=reaction_adapter(products),
                I=reaction_adapter(inhibitors))
        except Exception as e:
            self._manageExceptionReactionSystem(e)
            return

        if reaction in self.reaction_list:
            self.notify('Error: this reaction is already present')
            return
        
        self.reaction_list.append(reaction)

        self.listWidgetReactions_addReaction(reaction)
        self.notify('Added ' + str(reaction).replace('->', '⟶'))

        self.pushButtonCalculate_enable()
        self.actionSave.setEnabled(True)

    def pushButtonAdd_enable(self, string=None):
        if self.lineEditReactants.text() != '' and self.lineEditProducts.text() != '':
            self.pushButtonAdd.setEnabled(True)
        else:
            self.pushButtonAdd.setEnabled(False)


    def pushButtonDelete_clicked(self):
        i = 0
        while self.listWidgetReactions._checked_item_number:
            if self.listWidgetReactions.item(i).checkState():
                item = self.listWidgetReactions.takeItem(i)
                self.reaction_list.remove(Reaction(reaction_adapter(item.text()).replace('⟶', '->')))
                self.listWidgetReactions._checked_item_number -= 1
            else: i += 1
        self.pushButtonDelete.setEnabled(False)

        self.pushButtonCalculate_enable()
        self.actionSave.setEnabled(True)


    def pushButtonCalculate_clicked(self):
        if not self.pushButtonCalculate_generateContextPropertieSet():
            return

        DialogFBP(self,
            deepcopy(reaction_adapter(self.lineEditCalculatorSymbols.text())),
            deepcopy(self.spinBoxCalculatorSteps.value()),
            ReactionSet(deepcopy(self.reaction_list)),
            deepcopy(self.context_given_set[0]),
            deepcopy(self.context_given_set[1])
        ).show()

    def pushButtonCalculate_generateContextPropertieSet(self):
        self.context_given_set = [set(), set()]
        steps = self.spinBoxCalculatorSteps.value()

        for j in range(0, self.tableWidgetProperties.columnCount()):
            for i in range(0, len(self.context_given_set)):
                cellContent = reaction_adapter(
                    self.tableWidgetProperties.cellWidget(i, j).text())
                symbol_regex = Reaction._get_symbol_regex()

                symbol_list_regex = '^\s*(%?{}\s+)*%?{}\s*$'.format(symbol_regex, symbol_regex) # pylint: disable=W1401
                
                if (len(cellContent) and not cellContent.isspace() and
                        not re.match(symbol_list_regex, cellContent)):
                    self.notify('Error: invalid syntax in step {}'.format(str(j+1)))
                    return False
                
                pattern = '%{}'.format(symbol_regex)
                context_given_from = re.findall(pattern, cellContent)
                context_given = re.sub(pattern, '', cellContent)
                context_given = re.findall(symbol_regex, context_given)
                
                context_given = set(map(lambda s: (j, s), context_given))
                context_given_from = set(map(lambda s: s[1:], context_given_from))
                
                self.context_given_set[i] = self.context_given_set[i].union(context_given)

                for z in range(j, steps):
                    for symbol in context_given_from:
                        self.context_given_set[i].add((z, symbol))
                

            intersectionSet = self.context_given_set[0].intersection(self.context_given_set[1])
            if len(intersectionSet):
                symbols_set = sorted(list(set(map(lambda t: t[1], intersectionSet))))
                self.notify('Error in context properties step {}: {}'.format(str(j+1), ' '.join(symbols_set)))
                return False
        
        return True

    def pushButtonCalculate_enable(self, string=None):
        if self.lineEditCalculatorSymbols.text() != '' \
                and self.spinBoxCalculatorSteps.value() != 0 \
                and len(self.reaction_list):
            self.pushButtonCalculate.setEnabled(True)

            self.tableWidgetProperties.setEnabled(True)
            self.tableWidgetProperties_initialize()

        else:
            self.pushButtonCalculate.setEnabled(False)
            self.tableWidgetProperties.setEnabled(False)



    def listWidgetReactions_addReaction(self, reaction):
        item = QtWidgets.QListWidgetItem(reaction_invadapter(str(reaction)).replace('->', '⟶'))
        item.setFlags(item.flags() | QtCore.Qt.ItemIsUserCheckable)
        item.setCheckState(QtCore.Qt.Unchecked)
        self.listWidgetReactions.addItem(item)
    
    def listWidgetReactions_itemChanged(self, item):
        if item.checkState():
            self.listWidgetReactions._checked_item_number += 1
            self.pushButtonDelete.setEnabled(True)
        else:
            self.listWidgetReactions._checked_item_number -= 1
            if not self.listWidgetReactions._checked_item_number:
                self.pushButtonDelete.setEnabled(False)
    
    def listWidgetReactions_clear(self):
        self.listWidgetReactions.clear()
        self.listWidgetReactions._checked_item_number = 0
        self.pushButtonDelete.setEnabled(False)
        self.actionSave.setEnabled(False)
        self.pushButtonCalculate_enable()


    def tableWidgetProperties_initialize(self):
        if self.tableWidgetProperties.columnCount() > self.spinBoxCalculatorSteps.value():        
            self.tableWidgetProperties.setColumnCount(self.spinBoxCalculatorSteps.value())

        self.tableWidgetProperties_fillSpace()
        
        self.tableWidgetProperties.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Interactive)

    def tableWidgetProperties_fillSpace(self):
        if not self.tableWidgetProperties.isEnabled():
            return

        for _ in range(self.tableWidgetProperties.columnCount(), self.spinBoxCalculatorSteps.value()):
            scrollbar = self.tableWidgetProperties.horizontalScrollBar()
            if scrollbar.maximum() != 0 and scrollbar.maximum() != scrollbar.value() :
                break
            self.tableWidgetProperties_addColumn()
    
    def tableWidgetProperties_scrollBar_valueChanged(self, value):
        if not self.tableWidgetProperties.isEnabled():
            return

        if (self.tableWidgetProperties.columnCount() < self.spinBoxCalculatorSteps.value() and 
                value == self.tableWidgetProperties.horizontalScrollBar().maximum()):
            self.tableWidgetProperties_addColumn()


    def tableWidgetProperties_addColumn(self):
        column = self.tableWidgetProperties.columnCount()
        self.tableWidgetProperties.setColumnCount(column+1)

        self.tableWidgetProperties.setHorizontalHeaderItem(column, QtWidgets.QTableWidgetItem(str(column+1)))
        for i in range(0, self.tableWidgetProperties.rowCount()):
            cellWidget = self.tableWidgetProperties.cellWidget(i, column)
            if cellWidget == None:
                lineEdit = QtWidgets.QLineEdit()
                lineEdit.setValidator(self.validatorContextProperties)
                lineEdit.returnPressed.connect(self.pushButtonCalculate.click)
                lineEdit.setStatusTip("Example: A B %C D (%C means C from this step)")
                self.tableWidgetProperties.setCellWidget(i, column, lineEdit)
        # self.tableWidgetProperties.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)


    def statusbarChanged(self, string):
        if string and string[0:5] == 'Error':
            self.statusbar.setStyleSheet("QStatusBar{background:rgba(255,0,0,255);color:black;font-weight:bold;}")
        else:
            self.statusbar.setStyleSheet(self.statusbarStyle)


    def notify(self, message):
        self.statusbar.showMessage(message, msecs=4000)

        
    def _manageExceptionReactionSystem(self, e):
        if isinstance(e, ExceptionReactionSystem.InvalidSyntax):
            message = 'Error: symbols must be strings of letters'
        elif isinstance(e, ExceptionReactionSystem.ReactantSetCannotBeEmpty):
            message = 'Error: reactants cannot be empty'
        elif isinstance(e, ExceptionReactionSystem.ProductSetCannotBeEmpty):
            message = 'Error: products cannot be empty'
        elif isinstance(e, ExceptionReactionSystem.InvalidReaction):
            message = 'Error: this reaction is invalid'
        elif isinstance(e, ExceptionReactionSystem.InvalidReactionSet):
            message = 'Error: this reaction set is invalid'
        elif isinstance(e, ExceptionReactionSystem.InvalidNumber):
            message = 'Error: only natural numbers are accepted '
        elif isinstance(e, ExceptionReactionSystem.InvalidFormula):
            message = 'Error: formula invalid'
        else:
            message = 'Error'

        self.notify(message)
